Page 60,132
;===============================================================
;
;   eForth 1.0 by Bill Muench and C. H. Ting, 1990
;
;   This is an implementation for the NEC 78C10 microcomputer by 
;   John Talbert, 1994, Oberlin Conservatory.
;  
;   Register Use:       Interpreter Pointer = DE
;                       Data Stack Pointer  = SP
;                       Return Stack Pointer  = HL

;                       Free to use: BC, EA, VA, Alternate Registers.
;
;   'doList'  is accessed as a subroutine through a CALT instruction 
;            (Call to Jump Table).  This shows up as a 'DB 80H' line
;            in the $COLON and $USER Macros.  When executed the 
;            processor jumps to an address vector located at 80H.  The
;            vectored 'doList' code is then located at 0F0H. The word
;            'call,' was changed to load 80H into the code area for a
;            doLST assembly.
;
;   A 9600 Baud serial I/O is provided.  PortB/bit0 is used for serial 
;       output and PortC/bit3 (INT2) is used for the serial input.  The
;       serial input is interrupt driven with a vectored interrupt routine
;       located at 0A0H.  The code words ?RX, TX!, and !IO make up the 
;       rest of the serial I/O code.  Three USER variables have been set
;       up for use by these serial I/O routines:  SERIN, which holds the
;       received character and a flag; HAFBIT, which adjusts the software
;       timing of the receiver to read in the middle of each bit frame 
;       (set it for 1/2 the BITIME minus 5); and BITIME, which adjusts the 
;       software for a specific baud rate (17H for 9600 baud assuming a 12Mhz
;       processor clock).
;
;   The 78C10 is an 8-bit micro, therefore cell aligning to even addresses
;       is unnecessary. The $ALIGN Macro was taken out along with the NOP's
;       used for cell alignment in the other Macros.  All occurrences of the
;       word ALGND were erased also. The word SEE no longer works because it
;       relies on cell alignment.
;   
;   All of the system FORTH code is to be stored in ROM (up to 32K) starting
;       at address 0000H. Then there is 2K of RAM starting at address COOOH.
;       This memory setup required the following changes:
;               1) Return and Data stacks and TIB moved to RAM.
;                  (See the Memory allocation EQU assignments.)
;               2) The USER variables were moved to the micro's
;                  internal RAM at FF00H to FFFFH.
;               3) PAD word was changed to move the temporary buffer
;                  area to RAM space.
;               4) The vocabulary pointers found in the word FORTH were
;                  moved to RAM space by creating two new USER variables,
;                  FHEAD and FLINK and changing DOVOC to read:
;                  DW FHEAD,CNTXT,STORE,EXIT.
;               5) NTOP and CTOP were moved to RAM space to allow dictionary
;                  expansion into RAM space.
;
;   Several words were added to the ROM Dictionary.  The simple operators
;       1+,1-,2+,2-,2*,2/, were defined in machine code.  The words C, , 
;       CCOMPILE, CODE, and ENDCODE were created to enable the creation
;       of code definition.
;   
;    The NEC78C10 offers the following advantages:
;               1) Ten 16-bit internal registers and a 16-bit ALU. 
;                  Many 16-bit instructions for those FORTH stack operations.
;               2) Three 8-bit I/O ports.
;               3) Eight 8-bit Analog to Digital Converters.
;               4) Internal counters and programmable clock generators.
;               5) Internal hardware serial I/O.  (can be used for MIDI I/O).
;               6) 64K address space including 256 bytes of internal RAM.
;
;
;==================================================================
;; Version control
VER             EQU     01H                     ;major release version
EXT             EQU     01H                     ;minor extension

;; Constants
COMPO           EQU     040H                    ;lexicon compile only bit
IMEDD           EQU     080H                    ;lexicon immediate bit
MASKK           EQU     07F1FH                  ;lexicon bit mask

CELLL           EQU     2                       ;size of a cell
BASEE           EQU     10                      ;default radix
VOCSS           EQU     6                       ;depth of vocabulary stack

BKSPP           EQU     8                       ;backspace
LF              EQU     10                      ;line feed
CRR             EQU     13                      ;carriage return
ERR             EQU     27                      ;error escape
TIC             EQU     39                      ;tick

CALLL           EQU     80H                     ;CALT opcodes

;; Memory allocation    0//code>--//--<name//up>--<sp//tib>--rp//em

COLDD           EQU     00100H                  ;cold start
RPP             EQU     0C2F0H                  ;start of return stack (RP0)
TIBB            EQU     0C200H                  ;terminal input buffer (TIB)
SPP             EQU     0C1F0H                  ;start of data stack (SP0)
UPP             EQU     0FF00H                  ;start of user area (UP0)
NAMEE           EQU     01FF0H                  ;name dictionary
CODEE           EQU     00200H                  ;code dictionary
CTOP            EQU     0C400H                  ;RAM code dict. expansion
NTOP            EQU     0C7FFH                  ;RAM name dict. expansion
PADD            EQU     0C300H                  ;PAD area

;; Initialize assembly variables

_LINK   = 0                                     ;force a null link
_NAME   = NAMEE                                 ;initialize name pointer
_CODE   = CODEE                                 ;initialize code pointer
_USER   = 4*CELLL                               ;first user variable offset

;; Define assembly macros

;       Compile a code definition header.

$CODE   MACRO   LEX,NAME,LABEL
LABEL:                                          ;;assembly label
	_CODE   = $                             ;;save code pointer
	_LEN    = (LEX AND 01FH)/CELLL          ;;string cell count, round down
	_NAME   = _NAME-((_LEN+3)*CELLL)        ;;new header on cell boundary
ORG     _NAME                                   ;;set name pointer
	DW       _CODE,_LINK                    ;;token pointer and link
	_LINK   = $                             ;;link points to a name string
	DB      LEX,NAME                        ;;name string
ORG     _CODE                                   ;;restore code pointer
	ENDM

;       Compile a colon definition header.

$COLON  MACRO   LEX,NAME,LABEL
	$CODE   LEX,NAME,LABEL
	DB 80H                                  ;;include CALT doLIST
	ENDM

;       Compile a user variable header.

$USER   MACRO   LEX,NAME,LABEL
	$CODE   LEX,NAME,LABEL
	DB 80H                                  ;;include CALT doLIST
	DW      DOUSE,_USER                     ;;followed by doUSER and offset
	_USER   = _USER+CELLL                   ;;update user area offset
	ENDM

;       Compile an inline string.

D$      MACRO   FUNCT,STRNG
	DW      FUNCT                           ;;function
	_LEN    = $                             ;;save address of count byte
	DB      0,STRNG                         ;;count byte and string
	_CODE   = $                             ;;save code pointer
ORG     _LEN                                    ;;point to count byte
	DB      _CODE-_LEN-1                    ;;set count
ORG     _CODE                                   ;;restore code pointer
	ENDM

;       Assemble inline direct threaded code ending.

$NEXT   MACRO
	DB 48H,84H                              ;;EA<(DE)++,next code address into AX
	DB 48H,28H                              ;;JMP EA,jump directly to code address
	ENDM

;; Main entry points and COLD start data

MAIN    SEGMENT 
ASSUME  CS:MAIN,DS:MAIN,ES:MAIN,SS:MAIN

ORG     0000H

ORIG:   DB 54H,00,01,00                         ;RESET vector, JMP 0100H
	DB 0AAH,62H,0,0                         ;NMI vector, EI RETI
	DB 8 DUP(0)                             ;INT T0/T1 vector
	DB 54H,0A0H,00H, 5 DUP(0)               ;INT1/2 vector, JMP 00A0H
	DB 8 DUP(0)                             ;INT E1/E0 vector
	DB 8 DUP(0)                             ;INT EIN/AD vector
	DB 8 DUP(0)                             ;INT SR/ST vector
	DB 48 DUP(0)                            ;FREE
	DB 32 DUP(0)                            ;SOFTI vector at 0060H

ORG     00A0H

; Vectored INT2 routine for Serial Input from Host Computer.
; Uses address FFF0 as a counter location - do not use elsewhere!
	DB 0B1H                 ;PUSH BC
	DB 0B2H                 ;PUSH DE
	DB 0B0H                 ;PUSH VA
	DB 68H,0FFH             ;MVI, V<FF
	DB 71H,0F0H,07H         ;MVIW, (V/F0)<07, number of bits to receive.
	DB 70H,1FH,04CH,0FFH    ;LBCD, BC<(FF4C), wait for a half bit.
	DB 53H                  ;DCR, C<C-1 skip, LOOP1
	DB 0FEH                 ;JR, Jump to loop1
	DB 52H                  ;DCR, B<B-1 skip
	DB 0FCH                 ;JR, Jump to loop1
	DB 70H,1FH,4EH,0FFH     ;LBCD, BC<(FF4E),wait 1 bit time,  LOOP2
	DB 53H                  ;DCR, C<C-1 skip
	DB 0FEH                 ;JR, Jump to loop2
	DB 52H                  ;DCR, B<B-1 skip
	DB 0FCH                 ;JR, Jump to loop2
	DB 04CH,0C2H            ;MOV, A<PC, read serial input on pc3
	DB 48H,31H,48H,31H      ;Rotate PC3 bit into Cy
	DB 48H,31H,48H,31H      ;RLR, A rotate right 4xs
	DB 0CH                  ;MOV, A<D, D collects the bits 
	DB 48H,31H              ;RLR, shift in next bit, CY to top of D
	DB 1CH                  ;MOV, D<A
	DB 30H,0F0H             ;DCRW, (V/F0)<(V/F0)-1 skip
	DB 0E7H                 ;JR, Jump to loop2 for next bit.
	DB 70H,1FH,4EH,0FFH     ;LBCD, BC<(FF4E)
	DB 53H,0FEH,52H,0FCH    ;DCR JR DCR JR, stop bit loop time.
	DB 71H,04BH,0FFH        ;MVIW, (V/4B)<FF, load flag
	DB 0CH,63H,04AH         ;MOV STAW, A<D (V/4A)<A, load data
	DB 0A0H,0A2H,0A1H       ;POP, restore AV DE and BC
	DB 48H,44H,0            ;SKIT,NOP
	DB 0AAH,062H            ;EI RETI, enable interrupts and return

;; Kernel doLST routine.  Always accessed by the CALT instruction: 80H
;; which is a Call Subroutine to jump to address vector located at 0080H.

ORG     00F0H
	DB 33H,33H              ;HL<HL-2
	DB 0A6H                 ;EA<DE
	DB 48H,93H              ;(HL)<EA
	DB 0A2H                 ;POP DE previously pushed by CALT
	DB 48H,84H              ;EA<(DE)++, $NEXT
	DB 48H,28H              ;JMP EA
ORG     0080H
	DB 0F0H,0               ; set up vector to doLST

ORG     COLDD                   ;Beginning of Cold Boot
	DB 69H,0FH,4DH,0D0H     ;MM<0F, memory map (11-8)   
	DB 69H,0FFH,4DH,0D2H    ;MA<FF, pa inputs (4-2)
	DB 69H,00H,4DH,0D3H     ;MB<00, pb outputs (4-6)
	DB 64H,01H,05H          ;PB<5
	DB 4DH,0D7H             ;MF<00, pf outputs (4-15)
	DB 69H,0AH,4DH,0D4H     ;MC<0A, pc1/3 inputs (4-9)
	DB 69H,0BH,4DH,0D1H     ;MCC<0B, pc mode (4-8)
	DB 64H,02H,04H          ;PC<04
	DB 64H,81H,06H          ;SMH<06, serial mode (7-7)
	DB 69H,4EH,4DH,0CAH     ;SML<4E, serial mode (7-9)
	DB 04H                  ;SP<SPP, stack pointer=data stack
	DB LOW SPP
	DB HIGH SPP
	DB 34H                  ;HL<RPP, HL=return stack pointer
	DB LOW RPP
	DB HIGH RPP
	DB 69H,00H,4DH,0E8H     ;ZCM<0, zero cross disabled (3-26)
	DB 68H,0FFH             ;V<FF
	DB 10H,68H,0FFH,69H,0   ;V'<FF, A"<0, V<FF, A<0

;; timer setups for Midi and LCD use
	DB 69H,64H,4DH,0DAH     ;TM0<64, timer0 (5-1)
	DB 69H,0FFH,4DH,0DBH    ;TM1<FF, timer1 (5-1)
	DB 64H,85H,0B3H         ;TMM<B3, timer mode (5-6)
	DB 44H,60H,0EAH,48H,0D3H        ;ETM1<EA = EA60 (6-2)
	DB 64H,83H,0CCH         ;EOM<CC, timer event mode (6-14)
	DB 69H,5CH,4DH,0CCH     ;ETMM<5C, timer event mode (6-11)

	DB 54H,00,02H           ;JMP to 0200, high level cold start
				;COLD WORD MOVED TO THE START OF CODE AREA.             
				;ATTEMPTED TO AUTOMATE-JMP COLD-WITH $JUMP
				;BUT MACRO PRODUCES ERROR CODES.

; COLD start moves the following to USER variables.
; MUST BE IN SAME ORDER AS USER VARIABLES.

UZERO:          DW      4 DUP (0)               ;reserved
		DW      SPP                     ;SP0
		DW      RPP                     ;RP0
		DW      QRX                     ;'?KEY
		DW      TXSTO                   ;'EMIT
		DW      ACCEP                   ;'EXPECT
		DW      KTAP                    ;'TAP
		DW      TXSTO                   ;'ECHO
		DW      DOTOK                   ;'PROMPT
		DW      BASEE                   ;BASE
		DW      0                       ;tmp
		DW      0                       ;SPAN
		DW      0                       ;>IN
		DW      0                       ;#TIB
		DW      TIBB                    ;TIB
		DW      0                       ;CSP
		DW      INTER                   ;'EVAL
		DW      NUMBQ                   ;'NUMBER
		DW      0                       ;HLD
		DW      0                       ;HANDLER
		DW      0                       ;CONTEXT pointer
		DW      VOCSS DUP (0)           ;vocabulary stack
		DW      0                       ;CURRENT pointer
		DW      0                       ;vocabulary link pointer
		DW      0                       ;FORTH HEAD
		DW      0                       ;FORTH LINK
		DW      CTOP                    ;CP
		DW      NTOP                    ;NP
		DW      LASTN                   ;LAST
		DW      0                       ;SERIN host receive char & flag
		DW      06H                     ;HAFBIT time for serial host,
						; (1/2 BITIME - 5)
		DW      16H                     ;BITIME baud for serial host

ULAST:

ORG     CODEE                                   ;start code dictionary

;   COLD        ( -- )
;               The hilevel cold start sequence.
		CCOLD = $
		$COLON  4,'COLD',COLD
COLD1:          DW      DOLIT,UZERO,DOLIT,UPP
		DW      DOLIT,ULAST-UZERO,CMOVE ;initialize user area
		DW      PRESE                   ;initialize stack and TIB
		DW      TBOOT,ATEXE             ;application boot
		DW      FORTH,CNTXT,AT,DUPP     ;initialize search order
		DW      CRRNT,DSTOR,OVERT
		DW      QUIT                    ;start interpretation
		DW      BRAN,COLD1              ;just in case


;; Device dependent I/O

;   BYE         ( -- )
;               Exit eForth.
		$CODE   3,'BYE',BYE
		DB 54H,0,0                    ;JMP Reset Vector

;   ?RX         ( -- c T | F )
;               Return input character and true, or a false if no input.

		$CODE   3,'?RX',QRX
	DB 68H,0FFH             ;MVI, V<FF
	DB 01H,4BH              ;LDAW, A<(V/4B) read serial-in flag
	DB 47H,0FFH             ;ONI, A AND FF skip if flag not zero
	DB 0CAH                 ;JR, jump ahead1
	DB 71H,04BH,0           ;MVIW, (V/4B)<0, reset flag to zero
	DB 70H,1FH,4AH,0FFH     ;LBCD, BC<(FF4A), read serin data
	DB 0B1H                 ;PUSH BC, push serial input data to stack
	DB 69H,0FFH             ;A<FF
	DB 1BH                  ;C<A, AHEAD1
	DB 6AH,0                ;B<0
	DB 0B1H                 ;PUSH BC, push serial input flag to stack
	$NEXT


;   TX!         ( c -- )
;               Send character c to the output device.

		$CODE   3,'TX!',TXSTO
	DB 0BAH                 ;Disable Interrupts
	DB 0A1H                 ;POP BC, pop char into C
	DB 0B2H                 ;PUSH DE, store interpreter pointer
	DB 0BH,1CH              ;A<C, D<A, char in A and D
	DB 68H,0FFH             ;V<FF
	DB 71H,0F0H,07H         ;(V/F0)<7
	DB 60H,91H              ;A<A EXOR A
	DB 6DH,01H              ;E<01
	DB 4DH,0C1H             ;PB<A
	DB 70H,1FH,04EH,0FFH    ;BC<(FF4E) set baud, LOOP1
	DB 53H,0FEH,52H,0FCH    ;C<C-1, JR, B<B-1, JR, jr to loop1
	DB 0CH                  ;A<D
	DB 07H,01H              ;A<A AND 01
	DB 4DH,0C1H             ;PB<A, send a bit
	DB 0CH                  ;A<D
	DB 48H,31H              ;A rotate logical right
	DB 1CH                  ;D<A
	DB 0,0,0,0              ;NOPs to make rec loop = transmit loop.
	DB 30H,0F0H             ;(V/F0)<(V/F0)-1 skip
	DB 0E8H                 ;JR, jump to loop1
	DB 0DH                  ;A<E
	DB 51H                  ;A<A-1 skip
	DB 0C6H                 ;JR, jump to loop2
	DB 0A2H                 ;POP DE, restore interpreter pointer
	DB 0AAH                 ;Enable Interrupts
	$NEXT                   ;End of routine

	DB 6CH,03H              ;D<03, LOOP2
	DB 1DH                  ;E<A
	DB 71H,0F0H,01          ;(V/F0)<01
	DB 4FH,0D7H             ;JRE, jump to loop1

;   !IO         ( -- )
;               Initialize the serial I/O devices.

		$CODE   3,'!IO',STOIO
	DB 69H,0EFH,4DH,0C7H    ;MKL<EF, enable int2 interrupt and 
	DB 69H,0FFH,4DH,0C6H    ;MKH<FF, disable all others with mask
	DB 0AAH                 ;EI, enable interrupt
	$NEXT


;; The kernel

;   doLIT       ( -- w )
;               Push an inline literal.

		$CODE   COMPO+5,'doLIT',DOLIT
		DB 48H,84H                      ;EA<(DE)++
		DB 0B4H                         ;PUSH EA
		$NEXT

;   EXIT        ( -- )
;               Terminate a colon definition.

		$CODE   4,'EXIT',EXIT
		DB 48H,85H                      ;EA<(HL)++
		DB 0B6H                         ;DE<EA
		$NEXT

;   EXECUTE     ( ca -- )
;               Execute the word at ca.

		$CODE   7,'EXECUTE',EXECU
		DB 0A1H                         ;POP BC
		DB 21H                          ;JMP BC

;   next        ( -- )
;               Run time code for the single index loop.
;               : next ( -- ) \ hilevel model
;                 r> r> dup if 1 - >r @ >r exit then drop cell+ >r ;

		$CODE   COMPO+4,'next',DONXT
		DB 6AH,0                        ;B<00
		DB 6BH,1                        ;C<01
		DB 48H,83H                      ;EA<(HL)
		DB 74H,0B5H                     ;EA<EA-BC Skip if no borrow
		DB 0C9H                         ;JMP NEXT1
		DB 48H,93H                      ;(HL)<EA
		DB 48H,82H                      ;EA<(DE)
		DB 0B6H                         ;DE<EA
		$NEXT
NEXT1:          DB 22H,22H                      ;DE<DE+2
		DB 32H,32H                      ;HL<HL+2
		$NEXT

;   ?branch     ( f -- )
;               Branch if flag is zero.

		$CODE   COMPO+7,'?branch',QBRAN
		DB 6AH,0FFH                     ;B<FF
		DB 6BH,0FFH                     ;C<FF
		DB 0A4H                         ;POP EA
		DB 74H,0CDH                     ;EA AND BC Skip if not zero
		DB 0C6H                         ;JMP BRAN1
		DB 22H,22H                      ;DE<DE+2
		$NEXT
BRAN1:          DB 48H,82H                      ;EA<(DE)
		DB 0B6H                         ;DE<EA
		$NEXT

;   branch      ( -- )
;               Branch to an inline address.

		$CODE   COMPO+6,'branch',BRAN
		DB 48H,82H                      ;EA<(DE)
		DB 0B6H                         ;DE<EA
		$NEXT

;   !           ( w a -- )
;               Pop the data stack to memory.

		$CODE   1,'!',STORE
		DB 0A1H                         ;POP BC, address
		DB 0A4H                         ;POP EA, data
		DB 09H                          ;A<EAL
		DB 39H                          ;(BC)<A
		DB 12H                          ;BC<BC+1
		DB 08H                          ;A<EAH
		DB 39H                          ;(BC)<A
		$NEXT

;   @           ( a -- w )
;               Push data at memory location to the data stack.

		$CODE   1,'@',AT
		DB 0A1H                         ;POP BC
		DB 29H                          ;A<(BC)
		DB 19H                          ;EAL<A
		DB 12H                          ;BC<BC+1
		DB 29H                          ;A<(BC)
		DB 18H                          ;EAH<A
		DB 0B4H                         ;PUSH EA
		$NEXT

;   C!          ( c b -- )
;               Pop the data stack to byte memory.

		$CODE   2,'C!',CSTOR
		DB 0A1H                         ;POP BC address
		DB 0A4H                         ;POP AE data
		DB 09H                          ;A<EAL
		DB 39H                          ;(BC)<A
		$NEXT

;   C@          ( b -- c )
;               Push byte memory location to the data stack.

		$CODE   2,'C@',CAT
		DB 0A1H                         ;POP BC
		DB 29H                          ;A<(BC)
		DB 6AH,0                        ;B<00
		DB 1BH                          ;C<A
		DB 0B1H                         ;PUSH BC
		$NEXT

;   RP@         ( -- a )
;               Push the current RP to the data stack.

		$CODE   3,'RP@',RPAT
		DB 0B3H                         ;PUSH HL
		$NEXT

;   RP!         ( a -- )
;               Set the return stack pointer.

		$CODE   COMPO+3,'RP!',RPSTO
		DB 0A3H                         ;POP HL
		$NEXT

;   R>          ( -- w )
;               Pop the return stack to the data stack.

		$CODE   2,'R>',RFROM
		DB 48H,85H                      ;EA<(HL)++
		DB 0B4H                         ;PUSH EA
		$NEXT

;   R@          ( -- w )
;               Copy top of return stack to the data stack.

		$CODE   2,'R@',RAT
		DB 48H,83H                      ;EA<(HL)
		DB 0B4H                         ;PUSH EA
		$NEXT

;   >R          ( w -- )
;               Push the data stack to the return stack.

		$CODE   COMPO+2,'>R',TOR
		DB 33H,33H                      ;HL<HL-2
		DB 0A4H                         ;POP EA
		DB 48H,93H                      ;(HL)<EA
		$NEXT

;   SP@         ( -- a )
;               Push the current data stack pointer.

		$CODE   3,'SP@',SPAT
		DB 70H,0EH,0FEH,0FFH            ;(FFFE)<SP
		DB 70H,1FH,0FEH,0FFH            ;BC<(FFFE)
		DB 0B1H                         ;PUSH BC
		$NEXT

;   SP!         ( a -- )
;               Set the data stack pointer.

		$CODE   3,'SP!',SPSTO
		DB 0A1H                         ;POP BC
		DB 70H,1EH,0FEH,0FFH            ;(FFFE)<BC
		DB 70H,0FH,0FEH,0FFH            ;PC<(FFFE)
		$NEXT

;   DROP        ( w -- )
;               Discard top stack item.

		$CODE   4,'DROP',DROP
		DB 0A4H                         ;POP EA
		$NEXT

;   DUP         ( w -- w w )
;               Duplicate the top stack item.

		$CODE   3,'DUP',DUPP
		DB 0A4H                         ;POP EA
		DB 0B4H                         ;PUSH EA
		DB 0B4H                         ;PUSH EA
		$NEXT

;   SWAP        ( w1 w2 -- w2 w1 )
;               Exchange top two stack items.

		$CODE   4,'SWAP',SWAP
		DB 0A4H                         ;POP EA
		DB 0A1H                         ;POP BC
		DB 0B4H                         ;PUSH EA
		DB 0B1H                         ;PUSH BC
		$NEXT

;   OVER        ( w1 w2 -- w1 w2 w1 )
;               Copy second stack item to top.

		$CODE   4,'OVER',OVER
		DB 0A4H                         ;POP AE
		DB 0A1H                         ;POP BC
		DB 0B1H                         ;PUSH BC
		DB 0B4H                         ;PUSH AE
		DB 0B1H                         ;PUSH BC
		$NEXT

;   0<          ( n -- t )
;               Return true if n is negative.

		$CODE   2,'0<',ZLESS
		DB 0A1H                         ;POP BC
		DB 69H,0FFH                     ;A<FF
		DB 48H,06H                      ;B Shift Left, Skip if carry
		DB 69H,0                        ;A<00
		DB 1AH                          ;B<A
		DB 1BH                          ;C<A
		DB 0B1H                         ;PUSH BC
		$NEXT

;   AND         ( w w -- w )
;               Bitwise AND.

		$CODE   3,'AND',ANDD
		DB 0A1H                         ;POP BC
		DB 0A4H                         ;POP AE
		DB 74H,8DH                      ;EA<EA AND BC
		DB 0B4H                         ;PUSH EA
		$NEXT

;   OR          ( w w -- w )
;               Bitwise inclusive OR.

		$CODE   2,'OR',ORR
		DB 0A1H                         ;POP BC
		DB 0A4H                         ;POP EA
		DB 74H,9DH                      ;EA<EA OR BC
		DB 0B4H                          ;PUSH EA
		$NEXT

;   XOR         ( w w -- w )
;               Bitwise exclusive OR.

		$CODE   3,'XOR',XORR
		DB 0A1H                         ;POP BC
		DB 0A4H                         ;POP EA
		DB 74H,95H                      ;EA<EA EX-OR BC
		DB 0B4H                         ;PUSH EA
		$NEXT

;   UM+         ( w w -- w cy )
;               Add two numbers, return the sum and carry flag.

		$CODE   3,'UM+',UPLUS
		DB 0A1H                         ;POP BC
		DB 0A4H                         ;POP EA
		DB 69H,0                        ;A<00
		DB 74H,0A5H                     ;EA<EA+BC Skip if no carry
		DB 41H                          ;A<A+1
		DB 1BH                          ;C<A
		DB 6AH,0                        ;B<00
		DB 0B4H                         ;PUSH EA
		DB 0B1H                         ;PUSH BC
		$NEXT

;; System and user variables

;   doVAR       ( -- a )
;               Run time routine for VARIABLE and CREATE.

		$COLON  COMPO+5,'doVAR',DOVAR
		DW      RFROM,EXIT

;   UP          ( -- a )
;               Pointer to the user area.

		$COLON  2,'UP',UP
		DW      DOVAR
		DW      UPP

;   doUSER      ( -- a )
;               Run time routine for user variables.

		$COLON  COMPO+6,'doUSER',DOUSE
		DW      RFROM,AT,UP,AT,PLUS,EXIT

;   SP0         ( -- a )
;               Pointer to bottom of the data stack.

		$USER   3,'SP0',SZERO

;   RP0         ( -- a )
;               Pointer to bottom of the return stack.

		$USER   3,'RP0',RZERO

;   '?KEY       ( -- a )
;               Execution vector of ?KEY.

		$USER   5,"'?KEY",TQKEY

;   'EMIT       ( -- a )
;               Execution vector of EMIT.

		$USER   5,"'EMIT",TEMIT

;   'EXPECT     ( -- a )
;               Execution vector of EXPECT.

		$USER   7,"'EXPECT",TEXPE

;   'TAP        ( -- a )
;               Execution vector of TAP.

		$USER   4,"'TAP",TTAP

;   'ECHO       ( -- a )
;               Execution vector of ECHO.

		$USER   5,"'ECHO",TECHO

;   'PROMPT     ( -- a )
;               Execution vector of PROMPT.

		$USER   7,"'PROMPT",TPROM

;   BASE        ( -- a )
;               Storage of the radix base for numeric I/O.

		$USER   4,'BASE',BASE

;   tmp         ( -- a )
;               A temporary storage location used in parse and find.

		$USER   COMPO+3,'tmp',TEMP

;   SPAN        ( -- a )
;               Hold character count received by EXPECT.

		$USER   4,'SPAN',SPAN

;   >IN         ( -- a )
;               Hold the character pointer while parsing input stream.

		$USER   3,'>IN',INN

;   #TIB        ( -- a )
;               Hold the current count and address of the terminal input buffer.

		$USER   4,'#TIB',NTIB
		_USER = _USER+CELLL

;   CSP         ( -- a )
;               Hold the stack pointer for error checking.

		$USER   3,'CSP',CSP

;   'EVAL       ( -- a )
;               Execution vector of EVAL.

		$USER   5,"'EVAL",TEVAL

;   'NUMBER     ( -- a )
;               Execution vector of NUMBER?.

		$USER   7,"'NUMBER",TNUMB

;   HLD         ( -- a )
;               Hold a pointer in building a numeric output string.

		$USER   3,'HLD',HLD

;   HANDLER     ( -- a )
;               Hold the return stack pointer for error handling.

		$USER   7,'HANDLER',HANDL

;   CONTEXT     ( -- a )
;               A area to specify vocabulary search order.

		$USER   7,'CONTEXT',CNTXT
		_USER = _USER+VOCSS*CELLL       ;vocabulary stack

;   CURRENT     ( -- a )
;               Point to the vocabulary to be extended.

		$USER   7,'CURRENT',CRRNT
		_USER = _USER+CELLL             ;vocabulary link pointer

;   FHEAD       ( -- a )
;               Point to the FORTH vocab head pointer.
		$USER   5,'FHEAD',FHEAD

;   FLINK       ( -- a )
;               Point to the FORTH vocab link pointer.
		$USER   5,'FLINK',FLINK

;   CP          ( -- a )
;               Point to the top of the code dictionary.

		$USER   2,'CP',CP

;   NP          ( -- a )
;               Point to the bottom of the name dictionary.

		$USER   2,'NP',NP

;   LAST        ( -- a )
;               Point to the last name in the name dictionary.

		$USER   4,'LAST',LAST

;   SERIN       ( -- a )
;               Point to host serial input. Flag in high, char in low byte.

		$USER   5,'SERIN',SERIN

;   HAFBIT      ( -- a )      
;               Point to half bit time used by serial i/0 routines.

		$USER   6,'HAFBIT',HAFBIT

;   BITIME      ( -- a )
;               Point to bit time used to set serial i/o baud rate.

		$USER   6,'BITIME',BITIME

;; Common functions

;   doVOC       ( -- )
;               Run time action of VOCABULARY's.

		$COLON  COMPO+5,'doVOC',DOVOC
		DW      FHEAD,CNTXT,STORE,EXIT

;   FORTH       ( -- )
;               Make FORTH the context vocabulary.

		$COLON  5,'FORTH',FORTH
		DW      DOVOC,EXIT
		; Head and Link pointers normally here were moved to User Ram.

;   ?DUP        ( w -- w w | 0 )
;               Dup tos if its is not zero.

		$CODE  4,'?DUP',QDUP
		DB 6AH,0FFH             ;B<FF     
		DB 6BH,0FFH             ;C<FF
		DB 0A4H                 ;POP EA
		DB 74H,0DDH             ;EA AND BC, Skip if zero
		DB 0B4H                 ;PUSH EA
		DB 0B4H                 ;PUSH EA
		$NEXT

;   ROT         ( w1 w2 w3 -- w2 w3 w1 )
;               Rot 3rd item to top.

		$COLON  3,'ROT',ROT
		DW      TOR,SWAP,RFROM,SWAP,EXIT

;   2DROP       ( w w -- )
;               Discard two items on stack.

		$CODE  5,'2DROP',DDROP
		DB 0A4H,0A4H            ;POP EA, POP EA
		$NEXT

;   2DUP        ( w1 w2 -- w1 w2 w1 w2 )
;               Duplicate top two items.

		$CODE  4,'2DUP',DDUP
		DB 0A4H,0A1H            ;POP EA, POP BC
		DB 0B1H,0B4H            ;PUSH BC, PUSH EA
		DB 0B1H,0B4H            ;PUSH BC, PUSH EA
		$NEXT

;   +           ( w w -- sum )
;               Add top two items.

		$CODE  1,'+',PLUS
		DB 0A1H,0A4H            ;POP BC, POP EA
		DB 74H,0A5H             ;EA<EA+BC, Skip
		DB 0                    ;NOP
		DB 0B4H                 ;PUSH EA
		$NEXT

;   D+          ( d d -- d )
;               Double addition, as an example using UM+.
;
;               $COLON  2,'D+',DPLUS
;               DW      TOR,SWAP,TOR,UPLUS
;               DW      RFROM,RFROM,PLUS,PLUS,EXIT

;   NOT         ( w -- w )
;               One's complement of tos.

		$CODE  3,'NOT',INVER
		DB 0A1H                 ;POP BC
		DB 69H,0FFH             ;A<FF
		DB 60H,12H              ;B<B EX-OR A
		DB 60H,13H              ;C<C EX-OR A
		DB 0B1H                 ;PUSH BC
		$NEXT

;   NEGATE      ( n -- -n )
;               Two's complement of tos.

		$CODE  6,'NEGATE',NEGAT
		DB 0A1H                 ;POP BC
		DB 69H,0FFH             ;A<FF
		DB 60H,12H              ;B<B EX-OR A
		DB 60H,13H              ;C<C EX-OR A
		DB 12H                  ;BC<BC+1
		DB 0B1H                 ;PUSH BC
		$NEXT
		

;   DNEGATE     ( d -- -d )
;               Two's complement of top double.

		$COLON  7,'DNEGATE',DNEGA
		DW      INVER,TOR,INVER
		DW      DOLIT,1,UPLUS
		DW      RFROM,PLUS,EXIT

;   -           ( n1 n2 -- n1-n2 )
;               Subtraction.

		$CODE  1,'-',SUBB
		DB 0A1H                 ;POP BC
		DB 069H,0FFH            ;A<FF
		DB 060H,12H             ;B<B EX-OR A
		DB 060H,13H             ;C<C EX-OR A
		DB 12H                  ;BC<BC+1
		DB 0A4H                 ;POP EA
		DB 74H,0A5H             ;EA<EA+BC Skip
		DB 0                    ;NOP
		DB 0B4H                 ;PUSH EA
		$NEXT

;   ABS         ( n -- n )
;               Return the absolute value of n.

		$COLON  3,'ABS',ABSS
		DW      DUPP,ZLESS
		DW      QBRAN,ABS1
		DW      NEGAT
ABS1:           DW      EXIT

;   =           ( w w -- t )
;               Return true if top two are equal.

		$CODE  1,'=',EQUAL
		DB 0A4H,0A1H            ;POP EA, POP BC
		DB 69H,0FFH             ;A<FF
		DB 74H,0FDH             ;EA-BC, Skip if zero
		DB 69H,00H              ;A<00
		DB 1AH,1BH              ;B<A, C<A
		DB 0B1H                 ;PUSH BC
		$NEXT

;   U<          ( u u -- t )
;               Unsigned compare of top two items.

		$COLON  2,'U<',ULESS
		DW      DDUP,XORR,ZLESS
		DW      QBRAN,ULES1
		DW      SWAP,DROP,ZLESS,EXIT
ULES1:          DW      SUBB,ZLESS,EXIT

;   <           ( n1 n2 -- t )
;               Signed compare of top two items.

		$COLON  1,'<',LESS
		DW      DDUP,XORR,ZLESS
		DW      QBRAN,LESS1
		DW      DROP,ZLESS,EXIT
LESS1:          DW      SUBB,ZLESS,EXIT

;   MAX         ( n n -- n )
;               Return the greater of two top stack items.

		$CODE  3,'MAX',MAX
		DB 0A4H,0A1H            ;POP EA, POP BC
		DB 74H,0BDH             ;EA-BC, Skip if borrow  
		DB 0C2H                 ;Jump to Push EA
		DB 0B1H                 ;PUSH BC
		DB 0C1H                 ;Jump to next
		DB 0B4H                 ;PUSH EA
		$NEXT

;   MIN         ( n n -- n )
;               Return the smaller of top two stack items.

		$CODE  3,'MIN',MIN
		DB 0A4H,0A1H            ;POP EA, POP BC
		DB 74H,0BDH             ;EA-BC, Skip if borrow  
		DB 0C2H                 ;Jump to Push EA
		DB 0B4H                 ;PUSH EA
		DB 0C1H                 ;Jump to next
		DB 0B1H                 ;PUSH BC
		$NEXT

;   WITHIN      ( u ul uh -- t )
;               Return true if u is within the range of ul and uh.

		$COLON  6,'WITHIN',WITHI
		DW      OVER,SUBB,TOR                   ;ul <= u < uh
		DW      SUBB,RFROM,ULESS,EXIT

;; Quick Operators

;   1+          ( n -- n+1 )
		$CODE 2,'1+',ONEP
		DB 0A1H                 ;POP BC
		DB 12H                  ;BC<BC+1
		DB 0B1H                 ;PUSH BC
		$NEXT

;   1-          ( n -- n-1 )
		$CODE 2,'1-',ONEM
		DB 0A1H                 ;POP BC
		DB 013H                 ;BC<BC-1
		DB 0B1H                 ;PUSH BC
		$NEXT

;   2+          ( n -- n+2 )
		$CODE 2,'2+',TWOP
		DB 0A1H                 ;POP BC
		DB 12H,12H              ;BC<BC+2
		DB 0B1H                 ;PUSH BC
		$NEXT

;   2-          ( n -- n-2 )
		$CODE 2,'2-',TWOM
		DB 0A1H                 ;POP BC
		DB 13H,13H              ;BC<BC-2
		DB 0B1H                 ;PUSH BC
		$NEXT

;   2*          ( n -- n*2 )
		$CODE 2,'2*',TWOSL
		DB 0A4H                 ;POP EA
		DB 48H,0A4H             ;EA Logical Shift Left
		DB 0B4H                 ;PUSH EA
		$NEXT

;   2/          ( n -- n/2 )    
		$CODE 2,'2/',TWOSR
		DB 0A4H                 ;POP EA
		DB 48H,0A0H             ;EA Logical Shift Right
		DB 0B4H                 ;PUSH EA
		$NEXT

;; Divide

;   UM/MOD      ( udl udh u -- ur uq )
;               Unsigned divide of a double by a single. Return mod and quotient.

		$COLON  6,'UM/MOD',UMMOD
		DW      DDUP,ULESS
		DW      QBRAN,UMM4
		DW      NEGAT,DOLIT,15,TOR
UMM1:           DW      TOR,DUPP,UPLUS
		DW      TOR,TOR,DUPP,UPLUS
		DW      RFROM,PLUS,DUPP
		DW      RFROM,RAT,SWAP,TOR
		DW      UPLUS,RFROM,ORR
		DW      QBRAN,UMM2
		DW      TOR,DROP,ONEP,RFROM
		DW      BRAN,UMM3
UMM2:           DW      DROP
UMM3:           DW      RFROM
		DW      DONXT,UMM1
		DW      DROP,SWAP,EXIT
UMM4:           DW      DROP,DDROP
		DW      DOLIT,-1,DUPP,EXIT      ;overflow, return max

;   M/MOD       ( d n -- r q )
;               Signed floored divide of double by single. Return mod and quotient.

		$COLON  5,'M/MOD',MSMOD
		DW      DUPP,ZLESS,DUPP,TOR
		DW      QBRAN,MMOD1
		DW      NEGAT,TOR,DNEGA,RFROM
MMOD1:          DW      TOR,DUPP,ZLESS
		DW      QBRAN,MMOD2
		DW      RAT,PLUS
MMOD2:          DW      RFROM,UMMOD,RFROM
		DW      QBRAN,MMOD3
		DW      SWAP,NEGAT,SWAP
MMOD3:          DW      EXIT

;   /MOD        ( n n -- r q )
;               Signed divide. Return mod and quotient.

		$COLON  4,'/MOD',SLMOD
		DW      OVER,ZLESS,SWAP,MSMOD,EXIT

;   MOD         ( n n -- r )
;               Signed divide. Return mod only.

		$COLON  3,'MOD',MODD
		DW      SLMOD,DROP,EXIT

;   /           ( n n -- q )
;               Signed divide. Return quotient only.

		$COLON  1,'/',SLASH
		DW      SLMOD,SWAP,DROP,EXIT

;; Multiply

;   UM*         ( u u -- ud )
;               Unsigned multiply. Return double product.

		$COLON  3,'UM*',UMSTA
		DW      DOLIT,0,SWAP,DOLIT,15,TOR
UMST1:          DW      DUPP,UPLUS,TOR,TOR
		DW      DUPP,UPLUS,RFROM,PLUS,RFROM
		DW      QBRAN,UMST2
		DW      TOR,OVER,UPLUS,RFROM,PLUS
UMST2:          DW      DONXT,UMST1
		DW      ROT,DROP,EXIT

;   *           ( n n -- n )
;               Signed multiply. Return single product.

		$COLON  1,'*',STAR
		DW      UMSTA,DROP,EXIT

;   M*          ( n n -- d )
;               Signed multiply. Return double product.

		$COLON  2,'M*',MSTAR
		DW      DDUP,XORR,ZLESS,TOR
		DW      ABSS,SWAP,ABSS,UMSTA
		DW      RFROM
		DW      QBRAN,MSTA1
		DW      DNEGA
MSTA1:          DW      EXIT

;   */MOD       ( n1 n2 n3 -- r q )
;               Multiply n1 and n2, then divide by n3. Return mod and quotient.

		$COLON  5,'*/MOD',SSMOD
		DW      TOR,MSTAR,RFROM,MSMOD,EXIT

;   */          ( n1 n2 n3 -- q )
;               Multiply n1 by n2, then divide by n3. Return quotient only.

		$COLON  2,'*/',STASL
		DW      SSMOD,SWAP,DROP,EXIT

;; Miscellaneous

;   CELL+       ( a -- a )
;               Add cell size in byte to address.

		$COLON  5,'CELL+',CELLP
		DW      DOLIT,CELLL,PLUS,EXIT

;   CELL-       ( a -- a )
;               Subtract cell size in byte from address.

		$COLON  5,'CELL-',CELLM
		DW      DOLIT,0-CELLL,PLUS,EXIT

;   CELLS       ( n -- n )
;               Multiply tos by cell size in bytes.

		$COLON  5,'CELLS',CELLS
		DW      DOLIT,CELLL,STAR,EXIT

;   BL          ( -- 32 )
;               Return 32, the blank character.

		$COLON  2,'BL',BLANK
		DW      DOLIT,' ',EXIT

;   >CHAR       ( c -- c )
;               Filter non-printing characters.

		$COLON  5,'>CHAR',TCHAR
		DW      DOLIT,07FH,ANDD,DUPP    ;mask msb
		DW      DOLIT,127,BLANK,WITHI   ;check for printable
		DW      QBRAN,TCHA1
		DW      DROP,DOLIT,'_'          ;replace non-printables
TCHA1:          DW      EXIT

;   DEPTH       ( -- n )
;               Return the depth of the data stack.

		$COLON  5,'DEPTH',DEPTH
		DW      SPAT,SZERO,AT,SWAP,SUBB
		DW      DOLIT,CELLL,SLASH,EXIT

;   PICK        ( ... +n -- ... w )
;               Copy the nth stack item to tos.

		$COLON  4,'PICK',PICK
		DW      ONEP,CELLS
		DW      SPAT,PLUS,AT,EXIT

;; Memory access

;   +!          ( n a -- )
;               Add n to the contents at address a.

		$COLON  2,'+!',PSTOR
		DW      SWAP,OVER,AT,PLUS
		DW      SWAP,STORE,EXIT

;   2!          ( d a -- )
;               Store the double integer to address a.

		$COLON  2,'2!',DSTOR
		DW      SWAP,OVER,STORE
		DW      CELLP,STORE,EXIT

;   2@          ( a -- d )
;               Fetch double integer from address a.

		$COLON  2,'2@',DAT
		DW      DUPP,CELLP,AT
		DW      SWAP,AT,EXIT

;   COUNT       ( b -- b +n )
;               Return count byte of a string and add 1 to byte address.

		$COLON  5,'COUNT',COUNT
		DW      DUPP,ONEP
		DW      SWAP,CAT,EXIT

;   HERE        ( -- a )
;               Return the top of the code dictionary.

		$COLON  4,'HERE',HERE
		DW      CP,AT,EXIT

;   PAD         ( -- a )
;               Return the address of a temporary buffer.

		$COLON  3,'PAD',PAD
		DW      DOLIT,PADD,EXIT

;   TIB         ( -- a )
;               Return the address of the terminal input buffer.

		$COLON  3,'TIB',TIB
		DW      NTIB,CELLP,AT,EXIT

;   @EXECUTE    ( a -- )
;               Execute vector stored in address a.

		$COLON  8,'@EXECUTE',ATEXE
		DW      AT,QDUP                 ;?address or zero
		DW      QBRAN,EXE1
		DW      EXECU                   ;execute if non-zero
EXE1:           DW      EXIT                    ;do nothing if zero

;   CMOVE       ( b1 b2 u -- )
;               Copy u bytes from b1 to b2.

		$COLON  5,'CMOVE',CMOVE
		DW      TOR
		DW      BRAN,CMOV2
CMOV1:          DW      TOR,DUPP,CAT
		DW      RAT,CSTOR
		DW      ONEP
		DW      RFROM,ONEP
CMOV2:          DW      DONXT,CMOV1
		DW      DDROP,EXIT

;   FILL        ( b u c -- )
;               Fill u bytes of character c to area beginning at b.

		$COLON  4,'FILL',FILL
		DW      SWAP,TOR,SWAP
		DW      BRAN,FILL2
FILL1:          DW      DDUP,CSTOR,ONEP
FILL2:          DW      DONXT,FILL1
		DW      DDROP,EXIT

;   -TRAILING   ( b u -- b u )
;               Adjust the count to eliminate trailing white space.

		$COLON  9,'-TRAILING',DTRAI
		DW      TOR
		DW      BRAN,DTRA2
DTRA1:          DW      BLANK,OVER,RAT,PLUS,CAT,LESS
		DW      QBRAN,DTRA2
		DW      RFROM,ONEP,EXIT           ;adjusted count
DTRA2:          DW      DONXT,DTRA1
		DW      DOLIT,0,EXIT            ;count=0

;   PACK$       ( b u a -- a )
;               Build a counted string with u characters from b. Null fill.

		$COLON  5,'PACK$',PACKS
		DW      DUPP,TOR          ;strings only on cell boundary
		DW      OVER,DUPP,DOLIT,0
		DW      DOLIT,CELLL,UMMOD,DROP  ;count mod cell
		DW      SUBB,OVER,PLUS
		DW      DOLIT,0,SWAP,STORE      ;null fill cell
		DW      DDUP,CSTOR,ONEP           ;save count
		DW      SWAP,CMOVE,RFROM,EXIT   ;move string

;; Numeric output, single precision

;   DIGIT       ( u -- c )
;               Convert digit u to a character.

		$COLON  5,'DIGIT',DIGIT
		DW      DOLIT,9,OVER,LESS
		DW      DOLIT,7,ANDD,PLUS
		DW      DOLIT,'0',PLUS,EXIT

;   EXTRACT     ( n base -- n c )
;               Extract the least significant digit from n.

		$COLON  7,'EXTRACT',EXTRC
		DW      DOLIT,0,SWAP,UMMOD
		DW      SWAP,DIGIT,EXIT

;   <#          ( -- )
;               Initiate the numeric output process.

		$COLON  2,'<#',BDIGS
		DW      PAD,HLD,STORE,EXIT

;   HOLD        ( c -- )
;               Insert a character into the numeric output string.

		$COLON  4,'HOLD',HOLD
		DW      HLD,AT,ONEM
		DW      DUPP,HLD,STORE,CSTOR,EXIT

;   #           ( u -- u )
;               Extract one digit from u and append the digit to output string.

		$COLON  1,'#',DIG
		DW      BASE,AT,EXTRC,HOLD,EXIT

;   #S          ( u -- 0 )
;               Convert u until all digits are added to the output string.

		$COLON  2,'#S',DIGS
DIGS1:          DW      DIG,DUPP
		DW      QBRAN,DIGS2
		DW      BRAN,DIGS1
DIGS2:          DW      EXIT

;   SIGN        ( n -- )
;               Add a minus sign to the numeric output string.

		$COLON  4,'SIGN',SIGN
		DW      ZLESS
		DW      QBRAN,SIGN1
		DW      DOLIT,'-',HOLD
SIGN1:          DW      EXIT

;   #>          ( w -- b u )
;               Prepare the output string to be TYPE'd.

		$COLON  2,'#>',EDIGS
		DW      DROP,HLD,AT
		DW      PAD,OVER,SUBB,EXIT

;   str         ( n -- b u )
;               Convert a signed integer to a numeric string.

		$COLON  3,'str',STR
		DW      DUPP,TOR,ABSS
		DW      BDIGS,DIGS,RFROM
		DW      SIGN,EDIGS,EXIT

;   HEX         ( -- )
;               Use radix 16 as base for numeric conversions.

		$COLON  3,'HEX',HEX
		DW      DOLIT,16,BASE,STORE,EXIT

;   DECIMAL     ( -- )
;               Use radix 10 as base for numeric conversions.

		$COLON  7,'DECIMAL',DECIM
		DW      DOLIT,10,BASE,STORE,EXIT

;; Numeric input, single precision

;   DIGIT?      ( c base -- u t )
;               Convert a character to its numeric value. A flag indicates success.

		$COLON  6,'DIGIT?',DIGTQ
		DW      TOR,DOLIT,'0',SUBB
		DW      DOLIT,9,OVER,LESS
		DW      QBRAN,DGTQ1
		DW      DOLIT,7,SUBB
		DW      DUPP,DOLIT,10,LESS,ORR
DGTQ1:          DW      DUPP,RFROM,ULESS,EXIT

;   NUMBER?     ( a -- n T | a F )
;               Convert a number string to integer. Push a flag on tos.

		$COLON  7,'NUMBER?',NUMBQ
		DW      BASE,AT,TOR,DOLIT,0,OVER,COUNT
		DW      OVER,CAT,DOLIT,'$',EQUAL
		DW      QBRAN,NUMQ1
		DW      HEX,SWAP,ONEP
		DW      SWAP,ONEM
NUMQ1:          DW      OVER,CAT,DOLIT,'-',EQUAL,TOR
		DW      SWAP,RAT,SUBB,SWAP,RAT,PLUS,QDUP
		DW      QBRAN,NUMQ6
		DW      ONEM,TOR
NUMQ2:          DW      DUPP,TOR,CAT,BASE,AT,DIGTQ
		DW      QBRAN,NUMQ4
		DW      SWAP,BASE,AT,STAR,PLUS,RFROM
		DW      ONEP
		DW      DONXT,NUMQ2
		DW      RAT,SWAP,DROP
		DW      QBRAN,NUMQ3
		DW      NEGAT
NUMQ3:          DW      SWAP
		DW      BRAN,NUMQ5
NUMQ4:          DW      RFROM,RFROM,DDROP,DDROP,DOLIT,0
NUMQ5:          DW      DUPP
NUMQ6:          DW      RFROM,DDROP
		DW      RFROM,BASE,STORE,EXIT

;; Basic I/O

;   ?KEY        ( -- c T | F )
;               Return input character and true, or a false if no input.

		$COLON  4,'?KEY',QKEY
		DW      TQKEY,ATEXE,EXIT

;   KEY         ( -- c )
;               Wait for and return an input character.

		$COLON  3,'KEY',KEY
KEY1:           DW      QKEY
		DW      QBRAN,KEY1
		DW      EXIT

;   EMIT        ( c -- )
;               Send a character to the output device.

		$COLON  4,'EMIT',EMIT
		DW      TEMIT,ATEXE,EXIT

;   NUF?        ( -- t )
;               Return false if no input, else pause and if CR return true.

		$COLON  4,'NUF?',NUFQ
		DW      QKEY,DUPP
		DW      QBRAN,NUFQ1
		DW      DDROP,KEY,DOLIT,CRR,EQUAL
NUFQ1:          DW      EXIT

;   PACE        ( -- )
;               Send a pace character for the file downloading process.

		$COLON  4,'PACE',PACE
		DW      DOLIT,11,EMIT,EXIT

;   SPACE       ( -- )
;               Send the blank character to the output device.

		$COLON  5,'SPACE',SPACE
		DW      BLANK,EMIT,EXIT

;   SPACES      ( +n -- )
;               Send n spaces to the output device.

		$COLON  6,'SPACES',SPACS
		DW      DOLIT,0,MAX,TOR
		DW      BRAN,CHAR2
CHAR1:          DW      SPACE
CHAR2:          DW      DONXT,CHAR1
		DW      EXIT

;   TYPE        ( b u -- )
;               Output u characters from b.

		$COLON  4,'TYPE',TYPEE
		DW      TOR
		DW      BRAN,TYPE2
TYPE1:          DW      DUPP,CAT,EMIT
		DW      ONEP
TYPE2:          DW      DONXT,TYPE1
		DW      DROP,EXIT

;   CR          ( -- )
;               Output a carriage return and a line feed.

		$COLON  2,'CR',CR
		DW      DOLIT,CRR,EMIT
		DW      DOLIT,LF,EMIT,EXIT

;   do$         ( -- a )
;               Return the address of a compiled string.

		$COLON  COMPO+3,'do$',DOSTR
		DW      RFROM,RAT,RFROM,COUNT,PLUS
		DW      TOR,SWAP,TOR,EXIT

;   $"|		( -- a )
;               Run time routine compiled by $". Return address of a compiled string.

		$COLON  COMPO+3,'$"|',STRQP
		DW      DOSTR,EXIT              ;force a call to do$

;   ."|		( -- )
;               Run time routine of ." . Output a compiled string.

		$COLON  COMPO+3,'."|',DOTQP
		DW      DOSTR,COUNT,TYPEE,EXIT

;   .R          ( n +n -- )
;               Display an integer in a field of n columns, right justified.

		$COLON  2,'.R',DOTR
		DW      TOR,STR,RFROM,OVER,SUBB
		DW      SPACS,TYPEE,EXIT

;   U.R         ( u +n -- )
;               Display an unsigned integer in n column, right justified.

		$COLON  3,'U.R',UDOTR
		DW      TOR,BDIGS,DIGS,EDIGS
		DW      RFROM,OVER,SUBB
		DW      SPACS,TYPEE,EXIT

;   U.          ( u -- )
;               Display an unsigned integer in free format.

		$COLON  2,'U.',UDOT
		DW      BDIGS,DIGS,EDIGS
		DW      SPACE,TYPEE,EXIT

;   .           ( w -- )
;               Display an integer in free format, preceeded by a space.

		$COLON  1,'.',DOT
		DW      BASE,AT,DOLIT,10,XORR   ;?decimal
		DW      QBRAN,DOT1
		DW      UDOT,EXIT               ;no, display unsigned
DOT1:           DW      STR,SPACE,TYPEE,EXIT    ;yes, display signed

;   ?           ( a -- )
;               Display the contents in a memory cell.

		$COLON  1,'?',QUEST
		DW      AT,DOT,EXIT

;; Parsing

;   parse       ( b u c -- b u delta ; <string> )
;               Scan string delimited by c. Return found string and its offset.

		$COLON  5,'parse',PARS
		DW      TEMP,STORE,OVER,TOR,DUPP
		DW      QBRAN,PARS8
		DW      ONEM,TEMP,AT,BLANK,EQUAL
		DW      QBRAN,PARS3
		DW      TOR
PARS1:          DW      BLANK,OVER,CAT          ;skip leading blanks ONLY
		DW      SUBB,ZLESS,INVER
		DW      QBRAN,PARS2
		DW      ONEP
		DW      DONXT,PARS1
		DW      RFROM,DROP,DOLIT,0,DUPP,EXIT
PARS2:          DW      RFROM
PARS3:          DW      OVER,SWAP
		DW      TOR
PARS4:          DW      TEMP,AT,OVER,CAT,SUBB   ;scan for delimiter
		DW      TEMP,AT,BLANK,EQUAL
		DW      QBRAN,PARS5
		DW      ZLESS
PARS5:          DW      QBRAN,PARS6
		DW      ONEP
		DW      DONXT,PARS4
		DW      DUPP,TOR
		DW      BRAN,PARS7
PARS6:          DW      RFROM,DROP,DUPP
		DW      ONEP,TOR
PARS7:          DW      OVER,SUBB
		DW      RFROM,RFROM,SUBB,EXIT
PARS8:          DW      OVER,RFROM,SUBB,EXIT

;   PARSE       ( c -- b u ; <string> )
;               Scan input stream and return counted string delimited by c.

		$COLON  5,'PARSE',PARSE
		DW      TOR,TIB,INN,AT,PLUS     ;current input buffer pointer
		DW      NTIB,AT,INN,AT,SUBB     ;remaining count
		DW      RFROM,PARS,INN,PSTOR,EXIT

;   .(          ( -- )
;               Output following string up to next ) .

		$COLON  IMEDD+2,'.(',DOTPR
		DW      DOLIT,')',PARSE,TYPEE,EXIT

;   (           ( -- )
;               Ignore following string up to next ) . A comment.

		$COLON  IMEDD+1,'(',PAREN
		DW      DOLIT,')',PARSE,DDROP,EXIT

;   \           ( -- )
;               Ignore following text till the end of line.

		$COLON  IMEDD+1,'\',BKSLA
		DW      NTIB,AT,INN,STORE,EXIT

;   CHAR        ( -- c )
;               Parse next word and return its first character.

		$COLON  4,'CHAR',CHAR
		DW      BLANK,PARSE,DROP,CAT,EXIT

;   TOKEN       ( -- a ; <string> )
;               Parse a word from input stream and copy it to name dictionary.

		$COLON  5,'TOKEN',TOKEN
		DW      BLANK,PARSE,DOLIT,31,MIN
		DW      NP,AT,OVER,SUBB,CELLM
		DW      PACKS,EXIT

;   WORD        ( c -- a ; <string> )
;               Parse a word from input stream and copy it to code dictionary.

		$COLON  4,'WORD',WORDD
		DW      PARSE,HERE,PACKS,EXIT

;; Dictionary search

;   NAME>       ( na -- ca )
;               Return a code address given a name address.

		$COLON  5,'NAME>',NAMET
		DW      CELLM,CELLM,AT,EXIT

;   SAME?       ( a a u -- a a f \ -0+ )
;               Compare u cells in two strings. Return 0 if identical.

		$COLON  5,'SAME?',SAMEQ
		DW      TOR
		DW      BRAN,SAME2
SAME1:          DW      OVER,RAT,CELLS,PLUS,AT
		DW      OVER,RAT,CELLS,PLUS,AT
		DW      SUBB,QDUP
		DW      QBRAN,SAME2
		DW      RFROM,DROP,EXIT         ;strings not equal
SAME2:          DW      DONXT,SAME1
		DW      DOLIT,0,EXIT            ;strings equal

;   find        ( a va -- ca na | a F )
;               Search a vocabulary for a string. Return ca and na if succeeded.

		$COLON  4,'find',FIND
		DW      SWAP,DUPP,CAT
		DW      DOLIT,CELLL,SLASH,TEMP,STORE
		DW      DUPP,AT,TOR,CELLP,SWAP
FIND1:          DW      AT,DUPP
		DW      QBRAN,FIND6
		DW      DUPP,AT,DOLIT,MASKK,ANDD,RAT,XORR
		DW      QBRAN,FIND2
		DW      CELLP,DOLIT,-1          ;true flag
		DW      BRAN,FIND3
FIND2:          DW      CELLP,TEMP,AT,SAMEQ
FIND3:          DW      BRAN,FIND4
FIND6:          DW      RFROM,DROP
		DW      SWAP,CELLM,SWAP,EXIT
FIND4:          DW      QBRAN,FIND5
		DW      CELLM,CELLM
		DW      BRAN,FIND1
FIND5:          DW      RFROM,DROP,SWAP,DROP
		DW      CELLM
		DW      DUPP,NAMET,SWAP,EXIT

;   NAME?       ( a -- ca na | a F )
;               Search all context vocabularies for a string.

		$COLON  5,'NAME?',NAMEQ
		DW      CNTXT,DUPP,DAT,XORR     ;?context=also
		DW      QBRAN,NAMQ1
		DW      CELLM                   ;no, start with context
NAMQ1:          DW      TOR
NAMQ2:          DW      RFROM,CELLP,DUPP,TOR    ;next in search order
		DW      AT,QDUP
		DW      QBRAN,NAMQ3
		DW      FIND,QDUP               ;search vocabulary
		DW      QBRAN,NAMQ2
		DW      RFROM,DROP,EXIT         ;found name
NAMQ3:          DW      RFROM,DROP              ;name not found
		DW      DOLIT,0,EXIT            ;false flag

;; Terminal response

;   ^H          ( bot eot cur -- bot eot cur )
;               Backup the cursor by one character.

		$COLON  2,'^H',BKSP
		DW      TOR,OVER,RFROM,SWAP,OVER,XORR
		DW      QBRAN,BACK1
		DW      DOLIT,BKSPP,TECHO,ATEXE,ONEM
		DW      BLANK,TECHO,ATEXE
		DW      DOLIT,BKSPP,TECHO,ATEXE
BACK1:          DW      EXIT

;   TAP         ( bot eot cur c -- bot eot cur )
;               Accept and echo the key stroke and bump the cursor.

		$COLON  3,'TAP',TAP
		DW      DUPP,TECHO,ATEXE
		DW      OVER,CSTOR,ONEP,EXIT

;   kTAP        ( bot eot cur c -- bot eot cur )
;               Process a key stroke, CR or backspace.

		$COLON  4,'kTAP',KTAP
		DW      DUPP,DOLIT,CRR,XORR
		DW      QBRAN,KTAP2
		DW      DOLIT,BKSPP,XORR
		DW      QBRAN,KTAP1
		DW      BLANK,TAP,EXIT
KTAP1:          DW      BKSP,EXIT
KTAP2:          DW      DROP,SWAP,DROP,DUPP,EXIT

;   accept      ( b u -- b u )
;               Accept characters to input buffer. Return with actual count.

		$COLON  6,'accept',ACCEP
		DW      OVER,PLUS,OVER
ACCP1:          DW      DDUP,XORR
		DW      QBRAN,ACCP4
		DW      KEY,DUPP
;               DW      BLANK,SUBB,DOLIT,95,ULESS
		DW      BLANK,DOLIT,127,WITHI
		DW      QBRAN,ACCP2
		DW      TAP
		DW      BRAN,ACCP3
ACCP2:          DW      TTAP,ATEXE
ACCP3:          DW      BRAN,ACCP1
ACCP4:          DW      DROP,OVER,SUBB,EXIT

;   EXPECT      ( b u -- )
;               Accept input stream and store count in SPAN.

		$COLON  6,'EXPECT',EXPEC
		DW      TEXPE,ATEXE,SPAN,STORE,DROP,EXIT

;   QUERY       ( -- )
;               Accept input stream to terminal input buffer.

		$COLON  5,'QUERY',QUERY
		DW      TIB,DOLIT,80,TEXPE,ATEXE,NTIB,STORE
		DW      DROP,DOLIT,0,INN,STORE,EXIT

;; Error handling

;   CATCH       ( ca -- 0 | err# )
;               Execute word at ca and set up an error frame for it.

		$COLON  5,'CATCH',CATCH
		DW      SPAT,TOR,HANDL,AT,TOR   ;save error frame
		DW      RPAT,HANDL,STORE,EXECU  ;execute
		DW      RFROM,HANDL,STORE       ;restore error frame
		DW      RFROM,DROP,DOLIT,0,EXIT ;no error

;   THROW       ( err# -- err# )
;               Reset system to current local error frame an update error flag.

		$COLON  5,'THROW',THROW
		DW      HANDL,AT,RPSTO          ;restore return stack
		DW      RFROM,HANDL,STORE       ;restore handler frame
		DW      RFROM,SWAP,TOR,SPSTO    ;restore data stack
		DW      DROP,RFROM,EXIT

;   NULL$       ( -- a )
;               Return address of a null string with zero count.

		$COLON  5,'NULL$',NULLS
		DW      DOVAR                   ;emulate CREATE
		DW      0
		DB      99,111,121,111,116,101

;   ABORT       ( -- )
;               Reset data stack and jump to QUIT.

		$COLON  5,'ABORT',ABORT
		DW      NULLS,THROW

;   abort"	( f -- )
;               Run time routine of ABORT" . Abort with a message.

		$COLON  COMPO+6,'abort"',ABORQ
		DW      QBRAN,ABOR1             ;text flag
		DW      DOSTR,THROW             ;pass error string
ABOR1:          DW      DOSTR,DROP,EXIT         ;drop error

;; The text interpreter

;   $INTERPRET  ( a -- )
;               Interpret a word. If failed, try to convert it to an integer.

		$COLON  10,'$INTERPRET',INTER
		DW      NAMEQ,QDUP              ;?defined
		DW      QBRAN,INTE1
		DW      AT,DOLIT,COMPO,ANDD     ;?compile only lexicon bits
		D$      ABORQ,' compile only'
		DW      EXECU,EXIT              ;execute defined word
INTE1:          DW      TNUMB,ATEXE             ;convert a number
		DW      QBRAN,INTE2
		DW      EXIT
INTE2:          DW      THROW                   ;error

;   [           ( -- )
;               Start the text interpreter.

		$COLON  IMEDD+1,'[',LBRAC
		DW      DOLIT,INTER,TEVAL,STORE,EXIT

;   .OK         ( -- )
;               Display 'ok' only while interpreting.

		$COLON  3,'.OK',DOTOK
		DW      DOLIT,INTER,TEVAL,AT,EQUAL
		DW      QBRAN,DOTO1
		D$      DOTQP,' ok'
DOTO1:          DW      CR,EXIT

;   ?STACK      ( -- )
;               Abort if the data stack underflows.

		$COLON  6,'?STACK',QSTAC
		DW      DEPTH,ZLESS             ;check only for underflow
		D$      ABORQ,' underflow'
		DW      EXIT

;   EVAL        ( -- )
;               Interpret the input stream.

		$COLON  4,'EVAL',EVAL
EVAL1:          DW      TOKEN,DUPP,CAT          ;?input stream empty
		DW      QBRAN,EVAL2
		DW      TEVAL,ATEXE,QSTAC       ;evaluate input, check stack
		DW      BRAN,EVAL1
EVAL2:          DW      DROP,TPROM,ATEXE,EXIT   ;prompt

;; Shell

;   PRESET      ( -- )
;               Reset data stack pointer and the terminal input buffer.

		$COLON  6,'PRESET',PRESE
		DW      SZERO,AT,SPSTO
		DW      DOLIT,TIBB,NTIB,CELLP,STORE,EXIT

;   xio         ( a a a -- )
;               Reset the I/O vectors 'EXPECT, 'TAP, 'ECHO and 'PROMPT.

		$COLON  COMPO+3,'xio',XIO
		DW      DOLIT,ACCEP,TEXPE,DSTOR
		DW      TECHO,DSTOR,EXIT

;   FILE        ( -- )
;               Select I/O vectors for file download.

		$COLON  4,'FILE',FILE
		DW      DOLIT,PACE,DOLIT,DROP
		DW      DOLIT,KTAP,XIO,EXIT

;   HAND        ( -- )
;               Select I/O vectors for terminal interface.

		$COLON  4,'HAND',HAND
		DW      DOLIT,DOTOK,DOLIT,EMIT
		DW      DOLIT,KTAP,XIO,EXIT

;   I/O         ( -- a )
;               Array to store default I/O vectors.

		$COLON  3,'I/O',ISLO
		DW      DOVAR                   ;emulate CREATE
		DW      QRX,TXSTO               ;default I/O vectors

;   CONSOLE     ( -- )
;               Initiate terminal interface.

		$COLON  7,'CONSOLE',CONSO
		DW      ISLO,DAT,TQKEY,DSTOR    ;restore default I/O device
		DW      HAND,EXIT               ;keyboard input

;   QUIT        ( -- )
;               Reset return stack pointer and start text interpreter.

		$COLON  4,'QUIT',QUIT
		DW      RZERO,AT,RPSTO          ;reset return stack pointer
QUIT1:          DW      LBRAC                   ;start interpretation
QUIT2:          DW      QUERY                   ;get input
		DW      DOLIT,EVAL,CATCH,QDUP   ;evaluate input
		DW      QBRAN,QUIT2             ;continue till error
		DW      TPROM,AT,SWAP           ;save input device
		DW      CONSO,NULLS,OVER,XORR   ;?display error message
		DW      QBRAN,QUIT3
		DW      SPACE,COUNT,TYPEE       ;error message
		D$      DOTQP,' ? '             ;error prompt
QUIT3:          DW      DOLIT,DOTOK,XORR        ;?file input
		DW      QBRAN,QUIT4
		DW      DOLIT,ERR,EMIT          ;file error, tell host
QUIT4:          DW      PRESE                   ;some cleanup
		DW      BRAN,QUIT1

;; The compiler

;   '           ( -- ca )
;               Search context vocabularies for the next word in input stream.

		$COLON  1,"'",TICK
		DW      TOKEN,NAMEQ             ;?defined
		DW      QBRAN,TICK1
		DW      EXIT                    ;yes, push code address
TICK1:          DW      THROW                   ;no, error

;   ALLOT       ( n -- )
;               Allocate n bytes to the code dictionary.

		$COLON  5,'ALLOT',ALLOT
		DW      CP,PSTOR,EXIT           ;adjust code pointer

;   ,           ( w -- )
;               Compile an integer into the code dictionary.

		$COLON  1,',',COMMA
		DW      HERE,DUPP,CELLP         ;cell boundary
		DW      CP,STORE,STORE,EXIT     ;adjust code pointer, compile

;   C,          ( b -- )
;               Compile a byte into the code dictionary

		$COLON  2,'C,',CCOMMA
		DW      HERE,DUPP,ONEP
		DW      CP,STORE,CSTOR,EXIT

;   [COMPILE]   ( -- ; <string> )
;               Compile the next immediate word into code dictionary.

		$COLON  IMEDD+9,'[COMPILE]',BCOMP
		DW      TICK,COMMA,EXIT

;   COMPILE     ( -- )
;               Compile the next address in colon list to code dictionary.

		$COLON  COMPO+7,'COMPILE',COMPI
		DW      RFROM,DUPP,AT,COMMA     ;compile address
		DW      CELLP,TOR,EXIT          ;adjust return address

;   LITERAL     ( w -- )
;               Compile tos to code dictionary as an integer literal.

		$COLON  IMEDD+7,'LITERAL',LITER
		DW      COMPI,DOLIT,COMMA,EXIT

;   $,"		( -- )
;               Compile a literal string up to next " .

		$COLON  3,'$,"',STRCQ
		DW      DOLIT,'"',WORDD		;move string to code dictionary
		DW      COUNT,PLUS        ;calculate aligned end of string
		DW      CP,STORE,EXIT           ;adjust the code pointer

;   RECURSE     ( -- )
;               Make the current word available for compilation.

		$COLON  IMEDD+7,'RECURSE',RECUR
		DW      LAST,AT,NAMET,COMMA,EXIT

;; Structures

;   FOR         ( -- a )
;               Start a FOR-NEXT loop structure in a colon definition.

		$COLON  IMEDD+3,'FOR',FOR
		DW      COMPI,TOR,HERE,EXIT

;   BEGIN       ( -- a )
;               Start an infinite or indefinite loop structure.

		$COLON  IMEDD+5,'BEGIN',BEGIN
		DW      HERE,EXIT

;   NEXT        ( a -- )
;               Terminate a FOR-NEXT loop structure.

		$COLON  IMEDD+4,'NEXT',NEXT
		DW      COMPI,DONXT,COMMA,EXIT

;   UNTIL       ( a -- )
;               Terminate a BEGIN-UNTIL indefinite loop structure.

		$COLON  IMEDD+5,'UNTIL',UNTIL
		DW      COMPI,QBRAN,COMMA,EXIT

;   AGAIN       ( a -- )
;               Terminate a BEGIN-AGAIN infinite loop structure.

		$COLON  IMEDD+5,'AGAIN',AGAIN
		DW      COMPI,BRAN,COMMA,EXIT

;   IF          ( -- A )
;               Begin a conditional branch structure.

		$COLON  IMEDD+2,'IF',IFF
		DW      COMPI,QBRAN,HERE
		DW      DOLIT,0,COMMA,EXIT

;   AHEAD       ( -- A )
;               Compile a forward branch instruction.

		$COLON  IMEDD+5,'AHEAD',AHEAD
		DW      COMPI,BRAN,HERE,DOLIT,0,COMMA,EXIT

;   REPEAT      ( A a -- )
;               Terminate a BEGIN-WHILE-REPEAT indefinite loop.

		$COLON  IMEDD+6,'REPEAT',REPEA
		DW      AGAIN,HERE,SWAP,STORE,EXIT

;   THEN        ( A -- )
;               Terminate a conditional branch structure.

		$COLON  IMEDD+4,'THEN',THENN
		DW      HERE,SWAP,STORE,EXIT

;   AFT         ( a -- a A )
;               Jump to THEN in a FOR-AFT-THEN-NEXT loop the first time through.

		$COLON  IMEDD+3,'AFT',AFT
		DW      DROP,AHEAD,BEGIN,SWAP,EXIT

;   ELSE        ( A -- A )
;               Start the false clause in an IF-ELSE-THEN structure.

		$COLON  IMEDD+4,'ELSE',ELSEE
		DW      AHEAD,SWAP,THENN,EXIT

;   WHILE       ( a -- A a )
;               Conditional branch out of a BEGIN-WHILE-REPEAT loop.

		$COLON  IMEDD+5,'WHILE',WHILE
		DW      IFF,SWAP,EXIT

;   ABORT"	( -- ; <string> )
;               Conditional abort with an error message.

		$COLON  IMEDD+6,'ABORT"',ABRTQ
		DW      COMPI,ABORQ,STRCQ,EXIT

;   $"		( -- ; <string> )
;               Compile an inline string literal.

		$COLON  IMEDD+2,'$"',STRQ
		DW      COMPI,STRQP,STRCQ,EXIT

;   ."		( -- ; <string> )
;               Compile an inline string literal to be typed out at run time.

		$COLON  IMEDD+2,'."',DOTQ
		DW      COMPI,DOTQP,STRCQ,EXIT

;; Name compiler

;   ?UNIQUE     ( a -- a )
;               Display a warning message if the word already exists.

		$COLON  7,'?UNIQUE',UNIQU
		DW      DUPP,NAMEQ              ;?name exists
		DW      QBRAN,UNIQ1             ;redefinitions are OK
		D$      DOTQP,' reDef '         ;but warn the user
		DW      OVER,COUNT,TYPEE        ;just in case its not planned
UNIQ1:          DW      DROP,EXIT

;   $,n         ( na -- )
;               Build a new dictionary name using the string at na.

		$COLON  3,'$,n',SNAME
		DW      DUPP,CAT                ;?null input
		DW      QBRAN,PNAM1
		DW      UNIQU                   ;?redefinition
		DW      DUPP,LAST,STORE         ;save na for vocabulary link
		DW      HERE,SWAP               ;align code address
		DW      CELLM                   ;link address
		DW      CRRNT,AT,AT,OVER,STORE
		DW      CELLM,DUPP,NP,STORE     ;adjust name pointer
		DW      STORE,EXIT              ;save code pointer
PNAM1:          D$      STRQP,' name'           ;null input
		DW      THROW

;; FORTH compiler

;   $COMPILE    ( a -- )
;               Compile next word to code dictionary as a token or literal.

		$COLON  8,'$COMPILE',SCOMP
		DW      NAMEQ,QDUP              ;?defined
		DW      QBRAN,SCOM2
		DW      AT,DOLIT,IMEDD,ANDD     ;?immediate
		DW      QBRAN,SCOM1
		DW      EXECU,EXIT              ;its immediate, execute
SCOM1:          DW      COMMA,EXIT              ;its not immediate, compile
SCOM2:          DW      TNUMB,ATEXE             ;try to convert to number
		DW      QBRAN,SCOM3
		DW      LITER,EXIT              ;compile number as integer
SCOM3:          DW      THROW                   ;error

;   CCOMPILE    ( a -- )
;               Compile next byte to code dictionary as machine code.

		$COLON  8,'CCOMPILE',CCOMP
		DW      NAMEQ,QDUP              ;?defined
		DW      QBRAN,CCOM2
		DW      AT,DOLIT,IMEDD,ANDD     ;?immediate
		DW      QBRAN,CCOM1
		DW      EXECU,EXIT              ;its immediate, execute
CCOM1:          DW      DROP,EXIT               ;its not immediate,drop
CCOM2:          DW      TNUMB,ATEXE             ;try to convert to number
		DW      QBRAN,CCOM3
		DW      CCOMMA,EXIT             ;compile as code byte
CCOM3:          DW      THROW                   ;error

;   OVERT       ( -- )
;               Link a new word into the current vocabulary.

		$COLON  5,'OVERT',OVERT
		DW      LAST,AT,CRRNT,AT,STORE,EXIT

;   ;           ( -- )
;               Terminate a colon definition.

		$COLON  IMEDD+COMPO+1,';',SEMIS
		DW      COMPI,EXIT,LBRAC,OVERT,EXIT

;   ]           ( -- )
;               Start compiling the words in the input stream.

		$COLON  1,']',RBRAC
		DW      DOLIT,SCOMP,TEVAL,STORE,EXIT

;   call,       ( ca -- )
;               Assemble a call instruction to doLST.

		$COLON  5,'call,',CALLC
		DW      DOLIT,CALLL,CCOMMA,EXIT  ;Direct Threaded Code

;   :           ( -- ; <string> )
;               Start a new colon definition using next word as its name.

		$COLON  1,':',COLON
		DW      TOKEN,SNAME
		DW      CALLC,RBRAC,EXIT

;   IMMEDIATE   ( -- )
;               Make the last compiled word an immediate word.

		$COLON  9,'IMMEDIATE',IMMED
		DW      DOLIT,IMEDD,LAST,AT,AT,ORR
		DW      LAST,AT,STORE,EXIT

;; Defining words

;   USER        ( u -- ; <string> )
;               Compile a new user variable.

		$COLON  4,'USER',USER
		DW      TOKEN,SNAME,OVERT,CALLC
		DW      COMPI,DOUSE,COMMA,EXIT

;   CREATE      ( -- ; <string> )
;               Compile a new array entry without allocating code space.

		$COLON  6,'CREATE',CREAT
		DW      TOKEN,SNAME,OVERT,CALLC
		DW      COMPI,DOVAR,EXIT

;   VARIABLE    ( -- ; <string> )
;               Compile a new variable initialized to 0.

		$COLON  8,'VARIABLE',VARIA
		DW      CREAT,DOLIT,0,COMMA,EXIT

;   CODE        ( -- )
;               Start a new code definition using next word as its name.

		$COLON  4,'CODE',CODE
		DW      TOKEN,SNAME             
		DW      DOLIT,CCOMP,TEVAL,STORE,EXIT

;   ENDCODE     ( -- )
;               Terminate a code definition

		$COLON  IMEDD+COMPO+7,'ENDCODE',ENDCD
		DW      DOLIT,48H,CCOMMA,DOLIT,84H,CCOMMA       ;$NEXT
		DW      DOLIT,48H,CCOMMA,DOLIT,28H,CCOMMA
		DW      LBRAC,OVERT,EXIT

;; Tools

;   _TYPE       ( b u -- )
;               Display a string. Filter non-printing characters.

		$COLON  5,'_TYPE',UTYPE
		DW      TOR                     ;start count down loop
		DW      BRAN,UTYP2              ;skip first pass
UTYP1:          DW      DUPP,CAT,TCHAR,EMIT     ;display only printable
		DW      ONEP                      ;increment address
UTYP2:          DW      DONXT,UTYP1             ;loop till done
		DW      DROP,EXIT

;   dm+         ( a u -- a )
;               Dump u bytes from , leaving a+u on the stack.

		$COLON  3,'dm+',DMP
		DW      OVER,DOLIT,4,UDOTR      ;display address
		DW      SPACE,TOR               ;start count down loop
		DW      BRAN,PDUM2              ;skip first pass
PDUM1:          DW      DUPP,CAT,DOLIT,3,UDOTR  ;display numeric data
		DW      ONEP                      ;increment address
PDUM2:          DW      DONXT,PDUM1             ;loop till done
		DW      EXIT

;   DUMP        ( a u -- )
;               Dump u bytes from a, in a formatted manner.

		$COLON  4,'DUMP',DUMP
		DW      BASE,AT,TOR,HEX         ;save radix, set hex
		DW      DOLIT,16,SLASH          ;change count to lines
		DW      TOR                     ;start count down loop
DUMP1:          DW      CR,DOLIT,16,DDUP,DMP    ;display numeric
		DW      ROT,ROT
		DW      SPACE,SPACE,UTYPE       ;display printable characters
		DW      NUFQ,INVER              ;user control
		DW      QBRAN,DUMP2
		DW      DONXT,DUMP1             ;loop till done
		DW      BRAN,DUMP3
DUMP2:          DW      RFROM,DROP              ;cleanup loop stack, early exit
DUMP3:          DW      DROP,RFROM,BASE,STORE   ;restore radix
		DW      EXIT

;   .S          ( ... -- ... )
;               Display the contents of the data stack.

		$COLON  2,'.S',DOTS
		DW      CR,DEPTH                ;stack depth
		DW      TOR                     ;start count down loop
		DW      BRAN,DOTS2              ;skip first pass
DOTS1:          DW      RAT,PICK,DOT            ;index stack, display contents
DOTS2:          DW      DONXT,DOTS1             ;loop till done
		D$      DOTQP,' <sp'
		DW      EXIT

;   !CSP        ( -- )
;               Save stack pointer in CSP for error checking.

		$COLON  4,'!CSP',STCSP
		DW      SPAT,CSP,STORE,EXIT     ;save pointer

;   ?CSP        ( -- )
;               Abort if stack pointer differs from that saved in CSP.

		$COLON  4,'?CSP',QCSP
		DW      SPAT,CSP,AT,XORR        ;compare pointers
		D$      ABORQ,'stacks'          ;abort if different
		DW      EXIT

;   >NAME       ( ca -- na | F )
;               Convert code address to a name address.

		$COLON  5,'>NAME',TNAME
		DW      CRRNT                   ;vocabulary link
TNAM1:          DW      CELLP,AT,QDUP           ;check all vocabularies
		DW      QBRAN,TNAM4
		DW      DDUP
TNAM2:          DW      AT,DUPP                 ;?last word in a vocabulary
		DW      QBRAN,TNAM3
		DW      DDUP,NAMET,XORR         ;compare
		DW      QBRAN,TNAM3
		DW      CELLM                   ;continue with next word
		DW      BRAN,TNAM2
TNAM3:          DW      SWAP,DROP,QDUP
		DW      QBRAN,TNAM1
		DW      SWAP,DROP,SWAP,DROP,EXIT
TNAM4:          DW      DROP,DOLIT,0,EXIT       ;false flag

;   .ID         ( na -- )
;               Display the name at address.

		$COLON  3,'.ID',DOTID
		DW      QDUP                    ;if zero no name
		DW      QBRAN,DOTI1
		DW      COUNT,DOLIT,01FH,ANDD   ;mask lexicon bits
		DW      UTYPE,EXIT              ;display name string
DOTI1:          D$      DOTQP,' {noName}'
		DW      EXIT

;   WORDS       ( -- )
;               Display the names in the context vocabulary.

		$COLON  5,'WORDS',WORDS
		DW      CR,CNTXT,AT             ;only in context
WORS1:          DW      AT,QDUP                 ;?at end of list
		DW      QBRAN,WORS2
		DW      DUPP,SPACE,DOTID        ;display a name
		DW      CELLM,NUFQ              ;user control
		DW      QBRAN,WORS1
		DW      DROP
WORS2:          DW      EXIT

;; Hardware reset

;   VER         ( -- n )
;               Return the version number of this implementation.

		$COLON  3,'VER',VERSN
		DW      DOLIT,VER*256+EXT,EXIT

;   hi          ( -- )
;               Display the sign-on message of eForth.

		$COLON  2,'hi',HI
		DW      STOIO,CR                ;initialize I/O
		D$      DOTQP,'eForth v'   
		DW      BASE,AT,HEX
		DW      VERSN,BDIGS,DIG,DIG
		DW      DOLIT,'.',HOLD
		DW      DIGS,EDIGS,TYPEE
		DW      BASE,STORE,CR,EXIT

;   'BOOT       ( -- a )
;               The application startup vector.

		$COLON  5,"'BOOT",TBOOT
		DW      DOVAR
		DW      HI                      ;application to boot

;===============================================================

LASTN           EQU     _NAME+4                 ;last name address

NTOPP           EQU     _NAME-0      ;next available memory in ROM name dictionary
CTOPP           EQU     $+0          ;next available memory in ROM code dictionary
ROMSPC          EQU     NTOPP-CTOPP  ;UNUSED DICTIONARY ROM SPACE

MAIN    ENDS
END     ORIG

;===============================================================

